/*
 * Copyright (C) 2012-2013 The Dojo Foundation
 * 
 * Copyright (C) 2012-2013 Rex Young
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package com.fastmessenger.impl;

import java.lang.reflect.Method;
import java.util.LinkedList;

import com.fastmessenger.model.IReceiver;

class Dispatcher {

    protected Messenger messenger;

    protected boolean dispatcherRunning = true;
    protected LinkedList<PortRequest> requestQueue = new LinkedList<PortRequest> ();

    protected Head[] headcounts;

    protected Object lockObj = requestQueue;

    protected class Head {
        boolean headRunning = false;
        Object implObj;

        Head (Object implObj) {
            this.implObj = implObj;
        }
    }

    Dispatcher (Messenger messenger, Object[] implObjs) {
        this.messenger = messenger;

        headcounts = new Head[implObjs.length];
        for (int i = 0; i < implObjs.length; i++) {
            headcounts[i] = new Head (implObjs[i]);
        }
    }

    Method findServiceMethodIfAny (String port, Object... args) {

        if (headcounts[0].implObj instanceof IReceiver) {
            return null;
        } else {
            return Utils.getMethodByValues (headcounts[0].implObj.getClass (), port, args);
        }
    }

    void stop () {
        synchronized (lockObj) {
            dispatcherRunning = false;
        }
    }

    void dispatch (PortRequest request) {
        synchronized (lockObj) {
            if (dispatcherRunning) {
                requestQueue.add (request);
                launchOneIdleHead ();
            }
        }
    }

    private void launchOneIdleHead () {
        for (int i = 0; i < headcounts.length; i++) {
            if (!headcounts[i].headRunning) {

                final Head head = headcounts[i];
                messenger.getExecutorService ().submit (new Runnable () {
                    public void run () {
                        while (true) {
                            PortRequest request;
                            synchronized (lockObj) {
                                if (!dispatcherRunning || requestQueue.isEmpty ()) {
                                    head.headRunning = false;
                                    return;
                                }
                                request = requestQueue.poll ();
                            }
                            processPortRequest (head, request);
                        }
                    }
                });

                headcounts[i].headRunning = true;
                return;
            }
        }
    }

    protected void processPortRequest (Head head, PortRequest request) {
        synchronized (lockObj) {
            if (!dispatcherRunning) {
                return;
            }
        }

        Object returnValue;
        try {
            if (head.implObj instanceof IReceiver) {

                returnValue = ((IReceiver) head.implObj).onPortRequest (request.getPortAddress ()
                        .getPort (), request.getPortArguments ());
            } else {

                returnValue = request.getPortImplMethod ().invoke (head.implObj,
                        request.getPortArguments ());
            }
        } catch (Throwable e) {
            returnValue = e;
            e.printStackTrace (System.err);
        }

        if (request.getReturnAddress () != null) {
            try {
                if (request.getReturnValueIndex () >= 0) {
                    request.getReturnArguments ()[request.getReturnValueIndex ()] = returnValue;
                }

                messenger.placePortRequest (null, null, 0, request.getReturnAddress (),
                        request.getReturnArguments ());

            } catch (Throwable e) {
            }
        }
    }

}
