/*
 * Copyright (C) 2012-2013 The Dojo Foundation
 * 
 * Copyright (C) 2012-2013 Rex Young
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package com.fastmessenger.impl;

import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import com.fastmessenger.model.IMessenger;
import com.fastmessenger.model.IReceiver;
import com.fastmessenger.model.IReturn;

/**
 * A reference implementation for messenger object
 * 
 * @see IMessenger
 * 
 * @author Rex Young
 */
public class Messenger implements IMessenger {

    private ExecutorService executorService;

    private ReceiverManager receiverManager = new ReceiverManager ();
    private DispatcherManager dispatcherManager = new DispatcherManager ();
    private RouteManager routeManager = new RouteManager ();
    private ReplyManager replyManager = new ReplyManager ();

    /**
     * Constructor a messenger object with a fixed-size thread pool that is used to execute requests
     * for active objects.
     */
    public Messenger () {
        this (Executors.newFixedThreadPool (Runtime.getRuntime ().availableProcessors ()));
    }

    /**
     * Constructor a messenger object with a given ExecutorService that is used to execute requests
     * for active objects.
     * 
     * @param executorService
     *            a given ExecutorService
     */
    public Messenger (ExecutorService executorService) {
        this.executorService = executorService;

        registerReceiver (replyManager, replyManager.SelfAddress.getActiveobject (),
                replyManager.SelfAddress.getPort ());
    }

    /**
     * Returns the ExecutorService this messenger object is using
     * 
     * @return the ExecutorService this messenger object is using
     */
    public ExecutorService getExecutorService () {
        return executorService;
    }

    /**
     * @see IMessenger#registerReceiver(Object, String, String...)
     */
    synchronized public void registerReceiver (Object implObj, String receiver, String... ports) {

        registerReceiver (new Object[] { implObj }, receiver, ports);
    }

    /**
     * @see IMessenger#registerReceiver(Object[], String, String...)
     */
    synchronized public void registerReceiver (Object[] implObjs, String receiver, String... ports) {

        Dispatcher dispatcher = new Dispatcher (this, implObjs);

        registerReceiverImpl (receiver, ports, implObjs, dispatcher);
    }

    synchronized void registerReceiverImpl (String receiver, String[] ports, Object[] implObjs,
            Dispatcher dispatcher) {

        Utils.chkNonDupArray (implObjs, false);
        Utils.chkSameTypeArray (implObjs);
        Utils.chkNonEmptyId (receiver);
        Utils.chkNonDupArray (ports, true);
        if (!(implObjs[0] instanceof IReceiver)) {
            Utils.chkMethodNames (implObjs[0], ports);
        }

        if (receiverManager.getSortedPorts (receiver) == null) {

            receiverManager.addReceiver (receiver, ports);
            dispatcherManager.addDispatcher (receiver, dispatcher);
        } else {
            throw new IllegalArgumentException ("AppLogic error: receiverId already exist: " + receiver);
        }
    }

    /**
     * @see IMessenger#deregisterReceiver(String)
     */
    synchronized public void deregisterReceiver (String receiver) {

        Utils.chkNonEmptyId (receiver);

        if (receiverManager.getSortedPorts (receiver) != null) {

            receiverManager.removeReceiver (receiver);
            dispatcherManager.removeDispatcher (receiver);
        }
    }

    /**
     * @see IMessenger#registerRoute(String, String)
     */
    synchronized public void registerRoute (String senderPortQName, String receiverPortQName) {

        Address senderAddress = new Address (senderPortQName);
        Address receiverAddress = new Address (receiverPortQName);

        routeManager.addRoute (senderAddress, receiverAddress);
    }

    /**
     * @see IMessenger#deregisterRoute(String, String)
     */
    synchronized public void deregisterRoute (String senderPortQName, String receiverPortQName) {

        Address senderAddress = new Address (senderPortQName);
        Address receiverAddress = new Address (receiverPortQName);

        routeManager.removeRoute (senderAddress, receiverAddress);
    }

    /**
     * @see IMessenger#sendMessage(String, Object...)
     */
    synchronized public void sendMessage (String portQName, Object... args) {

        Address portAddress = new Address (portQName);

        placePortRequest (null, null, 0, portAddress, args);
    }

    /**
     * @see IMessenger#callService(String, Object...)
     */
    synchronized public <T> IReturn<T> callService (String portQName, Object... args) {

        Address portAddress = new Address (portQName);

        return placePortRequest (replyManager.SelfAddress, null, 0, portAddress, args);
    }

    /**
     * @see IMessenger#sendRequest(Object[], String, Object...)
     */
    synchronized public void sendRequest (Object[] listenerInfo, String portQName, Object... args) {

        Address portAddress = new Address (portQName);

        int returnValueIndex;
        Address returnAddress;
        Object[] returnArgs;

        if (listenerInfo[0] instanceof String) {
            returnValueIndex = -1;
            returnAddress = new Address ((String) listenerInfo[0]);
            returnArgs = new Object[listenerInfo.length - 1];
            System.arraycopy (listenerInfo, 1, returnArgs, 0, returnArgs.length);
        } else {
            returnValueIndex = ((Number) listenerInfo[0]).intValue ();
            if (returnValueIndex < 0 || returnValueIndex >= listenerInfo.length - 2) {
                throw new IllegalArgumentException ("AppLogic error: returnValueIndex is out of range: "
                        + returnValueIndex);
            }
            returnAddress = new Address ((String) listenerInfo[1]);
            returnArgs = new Object[listenerInfo.length - 2];
            System.arraycopy (listenerInfo, 2, returnArgs, 0, returnArgs.length);
        }

        placePortRequest (returnAddress, returnArgs, returnValueIndex, portAddress, args);
    }

    /**
     * @see IMessenger#publishMessage(String, Object...)
     */
    synchronized public void publishMessage (String senderPortQName, Object... args) {

        Address senderAddress = new Address (senderPortQName);

        List<Address> receiverAddressList = routeManager.getReceiversBySender (senderAddress);

        if (receiverAddressList != null) {
            for (Address receiverAddress : receiverAddressList) {
                try {
                    placePortRequest (null, null, 0, receiverAddress, args);
                } catch (RuntimeException e) {
                }
            }
        }
    }

    synchronized <T> IReturn<T> placePortRequest (Address returnAddress, Object[] returnArgs,
            int returnValueIndex, Address portAddress, Object[] args) {

        String[] sortedPorts;

        if (returnAddress != null && returnAddress != replyManager.SelfAddress) {

            sortedPorts = receiverManager.getSortedPorts (returnAddress.getActiveobject ());

            if (sortedPorts == null) {

                throw new IllegalArgumentException ("AppLogic error: returnId not found: " + returnAddress);

            } else if (returnAddress.getPort ().isEmpty ()) {

                returnAddress = new Address (returnAddress.getActiveobject (), sortedPorts[0]);

            } else if (Arrays.binarySearch (sortedPorts, returnAddress.getPort ()) < 0) {

                throw new IllegalArgumentException ("AppLogic error: returnService not found: "
                        + returnAddress);
            }
        }

        sortedPorts = receiverManager.getSortedPorts (portAddress.getActiveobject ());

        if (sortedPorts == null) {

            throw new IllegalArgumentException ("AppLogic error: receiverId not found: " + portAddress);

        } else if (portAddress.getPort ().isEmpty ()) {

            portAddress = new Address (portAddress.getActiveobject (), sortedPorts[0]);

        } else if (Arrays.binarySearch (sortedPorts, portAddress.getPort ()) < 0) {

            throw new IllegalArgumentException ("AppLogic error: receiverService not found: " + portAddress);
        }

        Method portImplMethod = dispatcherManager.findPortImplMethodIfAny (portAddress, args);

        PortRequest request = new PortRequest (returnAddress, returnArgs, returnValueIndex, portAddress,
                portImplMethod, args);

        if (returnAddress == null) {

            dispatcherManager.dispatchPortRequest (request);
            return null;

        } else {
            Reply<T> reply;

            if (returnAddress == replyManager.SelfAddress) {
                reply = new Reply<T> ();
                String replyId = replyManager.addReply (reply);
                reply.setReplyId (replyId);

                request.setReturnArguments (new Object[] { replyId, null });
                request.setReturnValueIndex (1);
            } else {
                reply = null;
            }

            dispatcherManager.dispatchPortRequest (request);

            return reply;
        }
    }

}
