/*
 * Copyright (C) 2012-2013 The Dojo Foundation
 * 
 * Copyright (C) 2012-2013 Rex Young
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package com.fastmessenger.impl;

import java.lang.reflect.Method;
import java.util.HashSet;

class Utils {

    static void chkNonNull (Object obj) {
        obj.getClass ();
    }

    static void chkNonEmptyId (String singleId) {
        if (singleId.isEmpty ()) {
            throw new IllegalArgumentException ("Validation error: empty Id found");
        } else {
            chkEmptyOkId (singleId);
        }
    }

    static void chkEmptyOkId (String singleId) {
        for (int i = 0; i < singleId.length (); i++) {
            char code = singleId.charAt (i);
            if (code == '*' || code == '?') {
                throw new IllegalArgumentException ("Validation error: wildcards not supported");
            } else if (code < 128 && !(code >= 'a' && code <= 'z') && !(code >= 'A' && code <= 'Z')
                    && !(code >= '0' && code <= '9') && code != '.' && code != '-' && code != '_') {
                throw new IllegalArgumentException ("Validation error: bad char found: " + code);
            }
        }
    }

    static void chkNonDupArray (Object[] objs, boolean isIdArray) {
        if (objs[0] == null) {
            throw new IllegalArgumentException ("Validation error: null element found at index 0");
        }

        if (isIdArray) {
            chkNonEmptyId ((String) objs[0]);
        }

        for (int i = 1; i < objs.length; i++) {
            if (objs[i] == null) {
                throw new IllegalArgumentException ("Validation error: null element found at " + i);
            }

            if (isIdArray) {
                chkNonEmptyId ((String) objs[i]);
            }

            for (int j = 0; j < i; j++) {
                if (objs[j] == objs[i] || objs[j].equals (objs[i])) {
                    throw new IllegalArgumentException ("Validation error: dup element found at " + i);
                }
            }
        }
    }

    static void chkSameTypeArray (Object[] objs) {
        for (int i = 1; i < objs.length; i++) {
            if (!objs[0].getClass ().isAssignableFrom (objs[i].getClass ())) {
                throw new IllegalArgumentException ("Validation error: incompatible element at " + i);
            }
        }
    }

    static void chkMethodNames (Object implObj, String[] methodNames) {
        HashSet<String> methodNameSet = new HashSet<String> ();
        for (Method method : implObj.getClass ().getMethods ()) {
            methodNameSet.add (method.getName ());
        }

        for (String name : methodNames) {
            if (!methodNameSet.contains (name)) {
                throw new IllegalArgumentException ("Validation error: method name not found: " + name);
            }
        }
    }

    static Method getMethodByValues (Class<?> implClass, String methodName, Object... methodArguments) {
        int nullValueCount = 0;
        Method method = null;

        Class<?>[] argValueClasses = new Class<?>[methodArguments.length];
        for (int i = 0; i < methodArguments.length; i++) {
            if (methodArguments[i] == null) {
                argValueClasses[i] = null;
                nullValueCount++;
            } else {
                argValueClasses[i] = methodArguments[i].getClass ();
            }
        }

        if (nullValueCount > 0) {
            method = getMethodByValueTypes (implClass, methodName, argValueClasses);
        } else {
            try {
                method = implClass.getMethod (methodName, argValueClasses);
            } catch (NoSuchMethodException e) {
                method = getMethodByValueTypes (implClass, methodName, argValueClasses);
            }
        }

        return method;
    }

    static Method getMethodByValueTypes (Class<?> clazz, String methodName, Class<?>... valueClasses) {
        for (Method method : clazz.getMethods ()) {
            if (method.getName ().equals (methodName)) {
                if (isAssignable (valueClasses, method.getParameterTypes ())) {
                    return method;
                }
            }
        }

        throw new IllegalArgumentException ("Validation error: method not found: " + clazz.getName () + "."
                + methodName + valueTypesToString (valueClasses));
    }

    static boolean isAssignable (Class<?>[] valueClasses, Class<?>[] paramClasses) {
        if (valueClasses.length != paramClasses.length) {
            return false;
        }

        for (int i = 0; i < valueClasses.length; i++) {
            Class<?> valueClass = valueClasses[i];
            Class<?> paramClass = paramClasses[i];

            if (valueClass == null) {
                if (paramClass.isPrimitive ()) {
                    return false;
                } else {
                    continue;
                }
            }

            if (paramClass.isPrimitive ()) {
                if (paramClass == Integer.TYPE)
                    paramClass = Integer.class;
                else if (paramClass == Long.TYPE)
                    paramClass = Long.class;
                else if (paramClass == Float.TYPE)
                    paramClass = Float.class;
                else if (paramClass == Double.TYPE)
                    paramClass = Double.class;
                else if (paramClass == Boolean.TYPE)
                    paramClass = Boolean.class;
                else if (paramClass == Character.TYPE)
                    paramClass = Character.class;
                else if (paramClass == Short.TYPE)
                    paramClass = Short.class;
                else if (paramClass == Byte.TYPE)
                    paramClass = Byte.class;
            }

            if (!paramClass.isAssignableFrom (valueClass)) {
                return false;
            }
        }

        return true;
    }

    static String valueTypesToString (Class<?>[] valueClasses) {
        StringBuilder buf = new StringBuilder (1024);
        buf.append ('(');
        for (int i = 0; i < valueClasses.length; i++) {
            if (i > 0) {
                buf.append (',');
            }
            buf.append ((valueClasses[i] == null) ? "null" : valueClasses[i].getName ());
        }
        buf.append (')');
        return buf.toString ();
    }

}
