/*
 * Copyright (C) 2012-2013 The Dojo Foundation
 * 
 * Copyright (C) 2012-2013 Rex Young
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package com.fastmessenger.model;

/**
 * A messenger object acts as a virtual machine that hosts active objects, and offers methods that
 * can be used to manage and call active objects.
 * <p>
 * An active object has a number of incoming and outgoing ports. The following example creates an
 * active object named <code>calculator</code> with incoming ports <code>multiply</code>,
 * <code>divide</code>, and <code>pow</code>.
 * <p>
 * <code>registerReceiver (implObj, "calculator", "multiply", "divide", "pow");</code>
 * <p>
 * In Java, incoming ports are viewed and implemented as methods while outgoing ports are not
 * managed. Each active object will be backed by a solid implementation object. In this example, the
 * implObj is such an implementation object for the <code>calculator</code>.
 * <p>
 * If the implObj is marked by IReceiver interface, the messenger object will forward requests at
 * all incoming ports to its IReceiver method.
 * <p>
 * <code>Object onPortRequest (String port, Object... reqArgs);</code>
 * <p>
 * If the implObj is not marked by IReceiver interface but it has same-name methods (e.g.
 * <code>multiply()</code>, <code>divide()</code>, and <code>pow()</code> methods) for all incoming
 * ports, the messenger object will forward requests at an incoming port to the corresponding
 * same-name method using reflection. Furthermore, the messenger object supports method
 * polymorphism. It will determine a method whose method signature is most appropriately matched by
 * the arguments in the request.
 * <p>
 * The messenger object guarantees the <code>implObj</code> and every implementation object a
 * single-threaded runtime environment. (Note: An implementation object exposed to other threads is
 * called a hybrid implementation object. Developers are responsible for handling multi-threading
 * inside hybrid objects.)
 * <p>
 * In order to call an incoming port on an active object, a caller must know their names and
 * arguments the port would accept. Names of active object and port can be specified by a single
 * argument <code>portQName</code> e.g. <code>"calculator:multiply"</code>. However parameters of
 * ports are not defined in the source code. Developers need to check documents about them somewhere
 * else. It means there is no static type checking.
 * <p>
 * The messenger object offers various methods to call active objects. All these methods are
 * non-blocking and will return immediately. And the target active object will run asynchronously.
 * Developers can always find a method that fits their programming style.
 * <p>
 * The <code>callService()</code> method is designed for developers who write codes in serial style.
 * After calling an active object, the caller will receive an <code>IReturn</code> object through
 * which the caller has choices either to wait for the result or do something else then come back
 * wait for the result. The caller may also choose to repeatedly query on (non-blocking) whether the
 * result is ready instead of being blocked while waiting.
 * <p>
 * Even if a specific port does not return a result value, this mechanism can be used to determine
 * when the port really finished the request.
 * <p>
 * The <code>sendRequest()</code> method is designed for developers who prefer event-driven
 * programming. The method submits a request in the same way as it does using the
 * <code>callService()</code>. Instead of receiving an <code>IReturn</code> object, the caller
 * provides a callback address so that the messenger object will notify the designated listener when
 * the port really finished the request.
 * <p>
 * The <code>sendMessage()</code> method is designed for developers who write codes in event-driven,
 * message-passing, or flow-based programming styles.
 * <p>
 * A messenger object does not manage outgoing ports for active objects. Outgoing ports are used as
 * source of published messages using the <code>publishMessage()</code> method. Messages published
 * are messages without targets. Their target active objects are determined by the messenger object
 * after messages have been published. The messenger object uses registered routes to find out
 * target active objects and forward messages to them using the <code>sendMessage()</code> method.
 * <p>
 * Outgoing ports are in a different namespace from the one incoming ports are in. It is possible to
 * register a route like:
 * <p>
 * <code>registerRoute("calculator:multiply", "calculator:multiply");</code>
 * <p>
 * A messenger object does not verify sender either. Different callers can publish messages with the
 * exactly same <code>senderPortQName</code>.
 * 
 * @author Rex Young
 * 
 * @see IReceiver
 * @see IReturn
 */
public interface IMessenger {

    /**
     * Places a request to invoke a port on an active object. The active object will execute the
     * request asynchronously. If there are other pending requests, the active object will execute
     * them one at a time in the order of arrival.
     * <p>
     * This method is not blocking and immediately returns an <code>IReturn</code> object that is
     * linked to the request. Using the <code>IReturn</code> object caller can either query whether
     * the request is completed or wait until the request is completed. And when the request is
     * completed, caller may retrieve a result value from the IReturn object if any.
     * 
     * @param portQName
     *            a qualified port name in the format of
     *            <code>"&lt;activeobject&gt;:&lt;port&gt;"</code>
     * @param args
     *            arguments used to invoke the port
     * @return an instance of IReturn that links to the request
     * 
     * @see IReturn
     */
    <T> IReturn<T> callService (String portQName, Object... args);

    /**
     * Places a request to invoke a port on an active object. The active object will execute the
     * request asynchronously. If there are other pending requests, the active object will execute
     * them one at a time in the order of arrival.
     * <p>
     * The listenerInfo is an array of objects:
     * <p>
     * <code>{ [n,] <code>"&lt;activeobject&gt;:&lt;port&gt;"</code> [, arg0, arg1, ...] }</code>
     * <p>
     * The qualified port name and pre-arranged arguments in the listenerInfo would be used to
     * invoke <code>sendMessage()</code> method by the messenger object when the request is
     * completed.
     * <p>
     * If the caller would like the result value of the request to be included in the pre-arranged
     * arguments, it needs to provide an optional <code>n</code> value which indicates the position,
     * starting from zero, in the arguments whose value would be substituted by the result value.
     * 
     * @param listenerInfo
     *            a listener which the messenger object would notify when the request is completed
     * @param portQName
     *            a qualified port name in the format of
     *            <code>"&lt;activeobject&gt;:&lt;port&gt;"</code>
     * @param args
     *            arguments used to invoke the port
     */
    void sendRequest (Object[] listenerInfo, String portQName, Object... args);

    /**
     * Places a request to invoke a port on an active object. The active object will execute the
     * request asynchronously. If there are other pending requests, the active object will execute
     * them one at a time in the order of arrival.
     * 
     * @param portQName
     *            a qualified port name in the format of
     *            <code>"&lt;activeobject&gt;:&lt;port&gt;"</code>
     * @param args
     *            arguments used to invoke the port
     */
    void sendMessage (String portQName, Object... args);

    /**
     * Publishes a message without knowing target receivers. This method will obtain a set of
     * receivers among registered routes whose sender matches the sender of the message. Then this
     * method forwards the message to each of these receivers by invoking <code>sendMessage()</code>
     * method.
     * <p>
     * This method does not verify whether the caller is the sender of the message as it claims to
     * be, and would conside it is normal when no receivers are found or a receiver does not exist.
     * 
     * @param senderPortQName
     *            sender's qualified port name in the format of
     *            <code>"&lt;activeobject&gt;:&lt;port&gt;"</code>
     * @param args
     *            sender's message in form of arguments
     * 
     * @see #registerRoute(String, String)
     * @see #deregisterRoute(String, String)
     */
    void publishMessage (String senderPortQName, Object... args);

    /**
     * Creates a receiver (namely an active object with incoming ports) and associates it with an
     * implementation object.
     * 
     * @param implObj
     *            an implementation object
     * @param receiver
     *            name of a receiver to be created
     * @param ports
     *            names of the receiver's incoming ports
     */
    void registerReceiver (Object implObj, String receiver, String... ports);

    /**
     * Creates a receiver (namely an active object with incoming ports) and associates it with a
     * number of implementation objects.
     * <p>
     * An active object with multiple implementation objects may be viewed as a group of unnamed
     * individual active objects. Requests placed on such an active object would be distributed
     * evenly to these implementation objects.
     * 
     * @param implObjs
     *            a number of implementation objects
     * @param receiver
     *            name of a receiver to be created
     * @param ports
     *            names of the receiver's incoming ports
     */
    void registerReceiver (Object[] implObjs, String receiver, String... ports);

    /**
     * Destroys a receiver.
     * 
     * @param receiver
     *            name of a receiver to be destroyed
     */
    void deregisterReceiver (String receiver);

    /**
     * Creates a route which contains a sender (as source) and a receiver (as target).
     * <p>
     * This method would not verify these qualified port names against existing receivers already
     * registered.
     * 
     * @param senderPortQName
     *            sender's qualified port name in the format of
     *            <code>"&lt;activeobject&gt;:&lt;port&gt;"</code>
     * @param receiverPortQName
     *            receiver's qualified port name in the format of
     *            <code>"&lt;activeobject&gt;:&lt;port&gt;"</code>
     */
    void registerRoute (String senderPortQName, String receiverPortQName);

    /**
     * Destroys a route.
     * 
     * @param senderPortQName
     *            sender's qualified port name in the format of
     *            <code>"&lt;activeobject&gt;:&lt;port&gt;"</code>
     * @param receiverPortQName
     *            receiver's qualified port name in the format of
     *            <code>"&lt;activeobject&gt;:&lt;port&gt;"</code>
     */
    void deregisterRoute (String senderPortQName, String receiverPortQName);

}
